# coding=utf-8
import logging
import os
import Pyro4

from pyage.core import address
from pyage.core.agent import   generate_agents, Agent
from pyage.core.locator import Pyro4Locator
from pyage.core.migration import  NoMigration
from pyage.core.statistics import SimpleStatistics
from pyage.core.stop_condition import StepLimitStopCondition
from pyage.solutions.evolution.crossover import  AverageFloatCrossover
from pyage.solutions.evolution.evaluation import  FloatRastriginEvaluation
from pyage.solutions.evolution.initializer import  FloatInitializer
from pyage.solutions.evolution.mutation import  UniformFloatMutation
from pyage.solutions.evolution.selection import TournamentSelection

logger = logging.getLogger(__name__)

agents_count = int(os.environ['AGENTS'])
logger.debug("EVO, %s agents", agents_count)

agents = generate_agents("agent", agents_count, Agent)

stop_condition = lambda: StepLimitStopCondition(500)

size = 500
operators = lambda: [FloatRastriginEvaluation(), TournamentSelection(size=125, tournament_size=125),
                     AverageFloatCrossover(size=size), UniformFloatMutation(probability=0.1, radius=1)]
initializer = lambda: FloatInitializer(500, size, -10, 10)

address_provider = address.SequenceAddressProvider

migration = NoMigration
locator = Pyro4Locator

ns_hostname = lambda: os.environ['NS_HOSTNAME']
pyro_daemon = Pyro4.Daemon()
daemon = lambda: pyro_daemon
stats = SimpleStatistics



import logging
import os
import urllib2
import time
import sys
from pyage.core.inject import InjectOptional, Inject

logger = logging.getLogger(__name__)


class Statistics(object):
    def update(self, step_count, agents):
        raise NotImplementedError()

    def summarize(self, agents):
        raise NotImplementedError()


class SimpleStatistics(Statistics):
    def __init__(self, plot_file_name='plot.png'):
        self.history = []
        self.plot_file_name = plot_file_name

    def update(self, step_count, agents):
        try:
            best_fitness = max(a.get_fitness() for a in agents)
            logger.info(best_fitness)
            self.history.append(best_fitness)
        except:
            logging.exception("")

    def summarize(self, agents):
        try:
            import pylab
            logger.debug(self.history)
            logger.debug("best genotype: %s", max(agents, key=lambda a: a.get_fitness).get_best_genotype())
            pylab.yscale('symlog')
            pylab.savefig(self.plot_file_name)
        except:
            logging.exception("")


class TimeStatistics(SimpleStatistics):
    @InjectOptional("notification_url")
    def __init__(self, plot_file_name='plot.png'):
        super(TimeStatistics, self).__init__(plot_file_name)
        self.times = []
        self.start = time.time()

    def update(self, step_count, agents):
        super(TimeStatistics, self).update(step_count, agents)
        try:
            self.times.append(time.time() - self.start)
        except:
            logging.exception("")

    def summarize(self, agents):
        try:
            import pylab
            pylab.plot(self.times, self.history)
            pylab.xlabel("time (s)")
            pylab.ylabel("fitness")
            pylab.yscale('symlog')
            pylab.savefig(self.plot_file_name)

            if hasattr(self, "notification_url"):
                url = self.notification_url + "?time=%s&agents=%s&conf=%s" % (
                    time.time() - self.start, os.environ['AGENTS'], sys.argv[1])
                logger.info(url)
                urllib2.urlopen(url)
            logger.debug(self.history)
            logger.debug("best genotype: %s", max(agents, key=lambda a: a.get_fitness).get_best_genotype())
        except:
            logging.exception("")


class NotificationStatistics(SimpleStatistics):
    @Inject("notification_url")
    def __init__(self):
        super(NotificationStatistics, self).__init__()
        self.start = time.time()

    def summarize(self, agents):
        try:
            url = self.notification_url + "?time=%s&agents=%s&conf=%s" % (
                time.time() - self.start, os.environ['AGENTS'], sys.argv[1])
            logger.info(url)
            urllib2.urlopen(url)

        except:
            logging.exception("")


class NoStatistics(Statistics):
    def update(self, step_count, agents):
        pass

    def summarize(self, agents):
        pass




# coding=utf-8
import logging
import os
import Pyro4

from pyage.core import address
from pyage.core.agent.agent import unnamed_agents
from pyage.core.agent.aggregate import AggregateAgent
from pyage.core.emas import EmasService
from pyage.core.locator import  RowLocator
from pyage.core.migration import Pyro4Migration
from pyage.core.statistics import  TimeStatistics
from pyage.core.stop_condition import StepLimitStopCondition
from pyage.solutions.evolution.crossover import   SinglePointCrossover
from pyage.solutions.evolution.evaluation import  FloatRastriginEvaluation
from pyage.solutions.evolution.initializer import  float_emas_initializer
from pyage.solutions.evolution.mutation import  UniformFloatMutation


logger = logging.getLogger(__name__)

agents_count = int(os.environ['AGENTS'])
logger.debug("EMAS, %s agents", agents_count)
agents = unnamed_agents(agents_count, AggregateAgent)

stop_condition = lambda: StepLimitStopCondition(100)

aggregated_agents = lambda: float_emas_initializer(40, energy=100, size=50, lowerbound=-10, upperbound=10)

emas = EmasService

minimal_energy = lambda: 0
reproduction_minimum = lambda: 90
migration_minimum = lambda: 120
newborn_energy = lambda: 100
transferred_energy = lambda: 40

evaluation = FloatRastriginEvaluation
crossover = SinglePointCrossover
mutation = lambda: UniformFloatMutation(probability=1, radius=1)

address_provider = address.SequenceAddressProvider

migration = Pyro4Migration
locator = RowLocator

ns_hostname = lambda: os.environ['NS_HOSTNAME']
pyro_daemon = Pyro4.Daemon()
daemon = lambda: pyro_daemon

stats = TimeStatistics


from datetime import datetime
import logging
import threading
from time import time
import sys

from pyage.core import inject

from pyage.core.workplace import Workplace


logger = logging.getLogger(__name__)

if __name__ == '__main__':
    start_time = time()
    level = logging.INFO
    if len(sys.argv) >= 3:
        level = getattr(logging, sys.argv[2].upper(), None)
    logging.basicConfig(filename='pyage-' + datetime.now().strftime("%Y%m%d_%H%M%S") + '.log', level=level)
    inject.config = sys.argv[1]
    logging.debug("config: %s", inject.config)
    workspace = Workplace()
    workspace.publish()
    workspace.publish_agents()
    logger.debug(workspace.address)
    if hasattr(workspace, "daemon"):
        thread = threading.Thread(target=workspace.daemon.requestLoop)
        thread.setDaemon(True)
        thread.start()
        import Pyro4
        Pyro4.config.COMMTIMEOUT = 1
    while not workspace.stopped:
        workspace.step()
    time = time() - start_time
    logger.debug("elapsed time: %s seconds", time)
    if hasattr(workspace, "daemon"):
        workspace.daemon.close()
    workspace.unregister_agents()
    workspace.unregister()


import os
import socket
from pyage.core.inject import Inject

class AddressProvider(object):
    def generate_address(self, obj):
        raise NotImplementedError()

counter = 0

class SequenceAddressProvider(AddressProvider):
    def __init__(self):
        super(SequenceAddressProvider, self).__init__()

    def generate_address(self, obj):
        global counter
        counter += 1
        return str(counter) + "." + socket.gethostname() + "." + str(os.getpid())

class Addressable(object):
    @Inject("address_provider")
    def __init__(self):
        super(Addressable, self).__init__()
        if hasattr(self, "name") and self.name:
            self.address = self.name + "." + socket.gethostname() + "." + str(os.getpid())
        else:
            self.address = self.address_provider.generate_address(self)

    def get_address(self):
        return self.address


import random
from pyage.core.operator import Operator
from pyage.solutions.evolution.genotype import PointGenotype, FloatGenotype

class AbstractCrossover(Operator):
    def __init__(self, type, size):
        super(AbstractCrossover, self).__init__(type)
        self.__size = size

    def process(self, population):
        parents = list(population)
        for i in range(len(population), self.__size):
            p1, p2 = random.sample(parents, 2)
            genotype = self.cross(p1, p2)
            population.append(genotype)


class AverageCrossover(AbstractCrossover):
    def __init__(self, size=100):
        super(AverageCrossover, self).__init__(PointGenotype, size)

    def cross(self, p1, p2):
        genotype = PointGenotype((p1.x + p2.x) / 2.0, (p1.y + p2.y) / 2.0)
        return genotype


class AverageFloatCrossover(AbstractCrossover):
    def __init__(self, size=100):
        super(AverageFloatCrossover, self).__init__(FloatGenotype, size)

    def cross(self, p1, p2):
        genotype = FloatGenotype([sum(p) / 2.0 for p in zip(p1.genes, p2.genes)])
        return genotype

class SinglePointCrossover(AbstractCrossover):
    def __init__(self, size=100):
        super(SinglePointCrossover, self).__init__(FloatGenotype, size)


    def cross(self, p1, p2):
        crossingPoint = random.randint(1, len(p1.genes))
        return FloatGenotype(p1.genes[:crossingPoint] + p2.genes[crossingPoint:])


import Pyro4

def query_property(workspace_name, agent_address, property):
    workspace = Pyro4.Proxy("PYRONAME:" + workspace_name)
    agent = workspace.get_agent(agent_address)
    return getattr(agent, property)



from math import cos, pi, sin, sqrt
from pyage.core.operator import Operator
from pyage.solutions.evolution.genotype import PointGenotype, FloatGenotype

A = 10

class FloatRastriginEvaluation(Operator):
    def __init__(self):
        super(FloatRastriginEvaluation, self).__init__(FloatGenotype)

    def process(self, population):
        for genotype in population:
            genotype.fitness = - self.__rastrigin(genotype.genes)

    def __rastrigin(self, genes):
        sum = len(genes) * A
        for gene in genes:
            sum += gene ** 2 - A * cos(2 * pi * gene)
        return sum


class RastriginEvaluation(Operator):
    def __init__(self):
        super(RastriginEvaluation, self).__init__(PointGenotype)

    def process(self, population):
        for genotype in population:
            genotype.fitness = - self.__rastrigin(genotype.x, genotype.y)

    def __rastrigin(self, x, y):
        return 2 * A + x ** 2 - A * cos(2 * pi * x) + y ** 2 - A * cos(2 * pi * y)


class DeJongEvaluation(Operator):
    def __init__(self, type=None):
        super(DeJongEvaluation, self).__init__(PointGenotype)

    def process(self, population):
        for genotype in population:
            genotype.fitness = - self.__DeJong(genotype.x, genotype.y)

    def __DeJong(self, x, y):
        return x ** 2 + y ** 2


class SchwefelEvaluation(Operator):
    def __init__(self):
        super(SchwefelEvaluation, self).__init__()

    def process(self, population):
        for genotype in population:
            genotype.fitness = - self.__schwefel(genotype.genes)

    def __schwefel(self, genes):
        sum = 418.9829
        for gene in genes:
            sum += -gene * sin(sqrt(abs(gene)))
        return sum




import logging
from pyage.core.inject import InjectOptional
from pyage.core.migration import  Pyro4Migration

logger = logging.getLogger(__name__)

class CountingPyro4Migration(Pyro4Migration):

    counter = 0

    @InjectOptional("migration_probability")
    def __init__(self):
        if hasattr(self, "migration_probability"):
            probability = self.migration_probability
        else:
            probability = 0.05
        super(CountingPyro4Migration, self).__init__(probability)

    def migrate(self, agent):
        if super(CountingPyro4Migration, self).migrate(agent):
            CountingPyro4Migration.counter += 1
          #  logging.info("migrated!")






import logging
import Pyro4
from pyage.core.inject import Inject
from pyage.core.workplace import WORKPLACE

class StopCondition(object):
    def should_stop(self, workplace):
        raise NotImplementedError()


class StepLimitStopCondition(StopCondition):
    def __init__(self, step_limit):
        super(StepLimitStopCondition, self).__init__()
        self.step_limit = step_limit

    def should_stop(self, workplace):
        return 0 < self.step_limit <= workplace.steps


class MinimumFitnessStopCondition(StopCondition):
    def __init__(self, minimal_fitness):
        super(MinimumFitnessStopCondition, self).__init__()
        self.minimal_fitness = minimal_fitness

    def should_stop(self, workplace):
        return workplace.get_fitness() >= self.minimal_fitness


class GlobalMinimumFitnessStopCondition(StopCondition):
    @Inject("ns_hostname")
    def __init__(self, minimal_fitness):
        super(GlobalMinimumFitnessStopCondition, self).__init__()
        self.minimal_fitness = minimal_fitness

    def should_stop(self, workplace):
        try:
            fitness = workplace.get_fitness()
            if fitness > workplace.get_best_known_fitness():
                ns = Pyro4.locateNS(self.ns_hostname)
                map(lambda w: Pyro4.Proxy(w).set_best_known_fitness(fitness), ns.list(WORKPLACE).values())
        except:
            logging.exception("")
        return workplace.get_best_known_fitness() >= self.minimal_fitness


import logging
import os
import urllib2
import Pyro4
import time
from datetime import datetime
from pyage.core.inject import Inject, InjectOptional
from pyage.core.statistics import Statistics
from pyage.core.workplace import WORKPLACE

logger = logging.getLogger(__name__)

class GlobalTimeStatistics(Statistics):
    @Inject("ns_hostname")
    @InjectOptional("notification_url")
    def __init__(self):
        super(GlobalTimeStatistics, self).__init__()
        self.fitness_output = open('fitness_pyage' + str(datetime.now()) + '.txt', 'a')
        self.fitness_output.write(
            '# DIMS:' + os.environ['DIMS'] + os.environ["MACHINES"] + ' ' + str(datetime.now()) + '\n')
        self.start = time.time()

    def update(self, step_count, agents):
        try:
            best_fitness = max(a.get_fitness() for a in agents)
            logger.info(best_fitness)
            if step_count % 100 == 0:
                ns = Pyro4.locateNS(self.ns_hostname)
                best_fitness = max(Pyro4.Proxy(w).get_fitness() for w in ns.list(WORKPLACE).values())
                self.append(best_fitness, step_count)
        except:
            logging.exception("")


    def append(self, best_fitness, step_count):
        self.fitness_output.write(str(time.time() - self.start) + ';' + str(abs(best_fitness)) + '\n')

    def summarize(self, agents):
        try:
            if hasattr(self, "notification_url"):
                url = self.notification_url + "?time=%s&agents=%s&conf=%s" % (
                    time.time() - self.start, os.environ['AGENTS'], os.environ['DIMS'])
                logger.info(url)
                urllib2.urlopen(url)

        except:
            logging.exception("")


class NotificationStats(Statistics):
    @Inject("notification_url")
    def __init__(self):
        super(NotificationStats, self).__init__()
        self.start = time.time()


    def update(self, step_count, agents):
        best_fitness = max(a.get_fitness() for a in agents)
        logger.info(best_fitness)

    def summarize(self, agents):
        try:
            if hasattr(self, "notification_url"):
                url = self.notification_url + "?time=%s&agents=%s&conf=%s" % (
                    time.time() - self.start, os.environ['AGENTS'], os.environ['DIMS'])
                logger.info(url)
                urllib2.urlopen(url)

        except:
            logging.exception("")


class MigrationNotificationStatistics(Statistics):
    @Inject("notification_url", "migration")
    def __init__(self):
        super(MigrationNotificationStatistics, self).__init__()
        self.start = time.time()


    def update(self, step_count, agents):
        best_fitness = max(a.get_fitness() for a in agents)
        logger.info(best_fitness)

    def summarize(self, agents):
        try:
            if hasattr(self, "notification_url"):
                url = self.notification_url + "?time=%s&agents=%s&conf=%s&pr=%s" % (
                    time.time() - self.start, os.environ['AGENTS'], str(self.migration.counter), str(os.environ['PR']))
                logger.info(url)
                urllib2.urlopen(url)

        except:
            logging.exception("")


import random
from pyage.core.operator import Operator

class TournamentSelection(Operator):
    def __init__(self, type=None, size=20, tournament_size=20):
        super(TournamentSelection, self).__init__()
        self.size = size
        self.tournament_size = tournament_size

    def process(self, population):
        p = list(population)
        population[:] = []
        for i in range(self.size):
            sample = random.sample(p, self.tournament_size)
            winner = max(sample, key=lambda genotype: genotype.fitness)
            population.append(winner)
            p.remove(winner)





# coding=utf-8
import logging
import os

from pyage.core import address

from pyage.core.agent.agent import unnamed_agents
from pyage.core.agent.aggregate import AggregateAgent
from pyage.core.emas import EmasService
from pyage.core.locator import GridLocator
from pyage.core.migration import ParentMigration
from pyage.core.stats.gnuplot import StepStatistics
from pyage.core.stop_condition import StepLimitStopCondition
from pyage.solutions.evolution.crossover import SinglePointCrossover
from pyage.solutions.evolution.evaluation import FloatRastriginEvaluation
from pyage.solutions.evolution.initializer import float_emas_initializer
from pyage.solutions.evolution.mutation import UniformFloatMutation


logger = logging.getLogger(__name__)

agents_count = int(os.environ['AGENTS'])
logger.debug("EMAS, %s agents", agents_count)
agents = unnamed_agents(agents_count, AggregateAgent)

stop_condition = lambda: StepLimitStopCondition(1000)

aggregated_agents = lambda: float_emas_initializer(40, energy=100, size=50, lowerbound=-10, upperbound=10)

emas = EmasService

minimal_energy = lambda: 0
reproduction_minimum = lambda: 90
migration_minimum = lambda: 120
newborn_energy = lambda: 100
transferred_energy = lambda: 40

evaluation = FloatRastriginEvaluation
crossover = SinglePointCrossover
mutation = lambda: UniformFloatMutation(probability=1, radius=1)

address_provider = address.SequenceAddressProvider

migration = ParentMigration
locator = GridLocator

stats = lambda: StepStatistics('fitness_%s_pyage.txt' % __name__)



import logging
from pyage.core.address import Addressable
from pyage.core.inject import Inject, InjectOptional
from pyage.core.agent.agent import AGENT
import signal

logger = logging.getLogger(__name__)

try:
    from Pyro4 import locateNS
except:
    print "Pyro4 not installed, running in local mode only"

WORKPLACE = "workplace"


class Workplace(Addressable):
    @Inject("agents:_Workplace__agents", "stop_condition", "stats")
    @InjectOptional("ns_hostname", "daemon")
    def __init__(self):
        super(Workplace, self).__init__()
        for agent in self.__agents.values():
            agent.parent = self
        self.steps = 0
        self.stopped = False
        self.best_known_fitness = -float("inf")

        def signal_handler(signal, frame):
            print 'You pressed Ctrl+C!'
            self.stop()

        signal.signal(signal.SIGINT, signal_handler)

    def get_agents(self):
        return self.__agents.values()

    def ping(self):
        return "pong"

    def stop(self):
        self.stopped = True
        self.stats.summarize(self.__agents.values())

    def step(self):
        self.steps += 1
        logger.info("=========STEP %s=============", self.steps)
        for agent in self.__agents.values():
            agent.step()
        self.stats.update(self.steps, self.__agents.values())
        if self.stop_condition.should_stop(self):
            self.stop()

    def get_fitness(self):
        return max(a.get_fitness() for a in self.get_agents())

    def get_best_known_fitness(self):
        return self.best_known_fitness

    def set_best_known_fitness(self, fitness):
        logger.info("new best fitness registered: " + str(fitness))
        self.best_known_fitness = fitness

    def publish_agents(self):
        for agent in self.__agents.values():
            try:
                uri = self.daemon.register(agent)
                ns = locateNS(self.ns_hostname)
                ns.register('%s.%s' % (AGENT, agent.address), uri)
                logger.debug(ns.list())
            except:
                logger.debug("could not locate nameserver")

    def unregister_agents(self):
        try:
            ns = locateNS(self.ns_hostname)
            for agent in self.__agents.values():
                ns.remove('%s.%s' % (AGENT, agent.address))
            logger.debug(ns.list())
        except:
            logger.debug("could not locate nameserver")

    def get_agent(self, address):
        return self.__agents[address]

    def remove_agent(self, address):
        agent = self.__agents[address]
        del self.__agents[address]
        agent.workspace = None
        return agent

    def publish(self):
        if hasattr(self, "ns_hostname") and hasattr(self, "daemon"):
            uri = self.daemon.register(self)
            try:
                ns = locateNS(self.ns_hostname)
                ns.register(WORKPLACE + '.' + self.address, uri)
                logger.debug(ns.list())
            except:
                logger.debug("could not locate nameserver")

    def unregister(self):
        try:
            ns = locateNS(self.ns_hostname)
            ns.remove(WORKPLACE + '.' + self.address)
            logger.debug(ns.list())
        except:
            logger.debug("could not locate nameserver")



class PointGenotype(object):
    def __init__(self, x, y):
        super(PointGenotype, self).__init__()
        self.x = x
        self.y = y
        self.fitness = None

    def __str__(self):
        return "(%s, %s), f:%s" % (self.x, self.y, self.fitness)

    def __repr__(self):
        return self.__str__()



class FloatGenotype(object):
    def __init__(self, genes):
        super(FloatGenotype, self).__init__()
        self.fitness = None
        self.genes = genes

    def __str__(self):
        return "%s, f:%s" % (self.genes, self.fitness)

    def __repr__(self):
        return self.__str__()




class Operator(object):
    def __init__(self, type=None):
        super(Operator, self).__init__()
        self.required_type = type

    def process(self, population):
        raise NotImplementedError()

    def is_compatible(self, operator):
        return  self.required_type == None\
                or operator.required_type == None\
                or operator.required_type == self.required_type


import logging
from pyage.core.address import Addressable
from pyage.core.inject import Inject

logger = logging.getLogger(__name__)


class EmasAgent(Addressable):
    @Inject("locator", "migration", "evaluation", "crossover", "mutation", "emas", "transferred_energy")
    def __init__(self, genotype, energy, name=None):
        self.name = name
        super(EmasAgent, self).__init__()
        self.genotype = genotype
        self.energy = energy
        self.steps = 0
        self.evaluation.process([genotype])

    def step(self):
        self.steps += 1
        try:
            neighbour = self.locator.get_neighbour(self)
            if neighbour:
                if self.emas.should_die(self):
                    self.death(neighbour)
                elif self.emas.should_reproduce(self, neighbour):
                    self.emas.reproduce(self, neighbour)
                else:
                    self.meet(neighbour)
                if self.emas.can_migrate(self):
                    self.migration.migrate(self)
        except:
            logging.exception('')

    def get_fitness(self):
        return self.genotype.fitness

    def get_best_genotype(self):
        return self.genotype

    def add_energy(self, energy):
        self.energy += energy

    def get_energy(self):
        return self.energy

    def get_genotype(self):
        return self.genotype

    def meet(self, neighbour):
        logger.debug(str(self) + "meets" + str(neighbour))
        if self.get_fitness() > neighbour.get_fitness():
            transfered_energy = min(self.transferred_energy, neighbour.energy)
            self.energy += transfered_energy
            neighbour.add_energy(-transfered_energy)
        elif self.get_fitness() < neighbour.get_fitness():
            transfered_energy = min(self.transferred_energy, self.energy)
            self.energy -= transfered_energy
            neighbour.add_energy(transfered_energy)

    def death(self, neighbour):
        self.distribute_energy()
        neighbour.add_energy(self.energy)
        self.energy = 0
        self.parent.remove_agent(self)
        logger.debug(str(self) + "died!")

    def distribute_energy(self):
        logger.debug("death, energy level: %d" % self.energy)
        if self.energy > 0:
            siblings = set(self.parent.get_agents())
            siblings.remove(self)
            portion = self.energy / len(siblings)
            if portion > 0:
                logger.debug("passing %d portion of energy to %d agents" % (portion, len(siblings)))
                for agent in siblings:
                    agent.add_energy(portion)
            left = self.energy % len(siblings)
            logger.debug("distributing %d left energy" % left)
            while left > 0:
                siblings.pop().add_energy(1)
                left -= 1


class EmasService(object):
    @Inject("minimal_energy", "reproduction_minimum", "migration_minimum", "newborn_energy")
    def __init__(self):
        super(EmasService, self).__init__()

    def should_die(self, agent):
        return agent.get_energy() <= self.minimal_energy

    def should_reproduce(self, a1, a2):
        return a1.get_energy() > self.reproduction_minimum and a2.get_energy() > self.reproduction_minimum

    def can_migrate(self, agent):
        return agent.get_energy() > self.migration_minimum and len(agent.parent.get_agents()) > 10

    def reproduce(self, a1, a2):
        logger.debug(str(a1) + " " + str(a2) + "reproducing!")
        energy = self.newborn_energy / 2 * 2
        a1.energy -= self.newborn_energy / 2
        a2.add_energy(-self.newborn_energy / 2)
        genotype = a1.crossover.cross(a1.genotype, a2.get_genotype())
        a1.mutation.mutate(genotype)
        a1.parent.add_agent(EmasAgent(genotype, energy))




import logging
import random
import Pyro4
from pyage.core.agent.agent import AGENT
from pyage.core.inject import Inject

logger = logging.getLogger(__name__)

class Migration(object):
    def migrate(self, agent):
        raise NotImplementedError()


class Pyro4Migration(Migration):
    @Inject("ns_hostname")
    def __init__(self, probability = 0.05):
        super(Pyro4Migration, self).__init__()
        self.probability = probability

    def migrate(self, agent):
        try:
            if random.random() < self.probability and len(agent.parent.get_agents()) > 1:
                logger.debug("migrating!")
                aggregate = self.__get_random_aggregate(agent)
                logger.debug(aggregate.get_address())
                aggregate.add_agent(agent.parent.remove_agent(agent))
                return True
        except:
            logging.exception("")
        return False

    def __get_random_aggregate(self, agent):
        ns = Pyro4.locateNS(self.ns_hostname)
        agents = ns.list(AGENT)
        logger.debug(agents)
        del agents[AGENT + "." + agent.parent.address]
        return Pyro4.Proxy(random.choice(agents.values()))


class ParentMigration(Migration):
    def migrate(self, agent):
        try:
            if random.random() > 0.95 and len(agent.parent.get_agents()) > 1 and len(
                agent.parent.parent.get_agents()) > 1:
                logger.debug("migrating!")
                aggregate = self.__get_random_aggregate(agent)
                logger.debug(aggregate.get_address())
                aggregate.add_agent(agent.parent.remove_agent(agent))
                return True
        except:
            logging.exception("")
        return False

    def __get_random_aggregate(self, agent):
        siblings = list(agent.parent.parent.get_agents())
        return random.choice(siblings)


class NoMigration(Migration):
    def migrate(self, agent):
        pass


import logging
import time
from pyage.core.statistics import Statistics

logger = logging.getLogger(__name__)


class StepStatistics(Statistics):
    def __init__(self, output_file_name='fitness_pyage.txt'):
        self.history = []
        self.fitness_output = open(output_file_name, 'a')

    def __del__(self):
        self.fitness_output.close()

    def append(self, best_fitness, step_count):
        self.fitness_output.write(str(step_count - 1) + ';' + str(abs(best_fitness)) + '\n')

    def update(self, step_count, agents):
        try:
            best_fitness = max(a.get_fitness() for a in agents)
            logger.info(best_fitness)
            self.history.append(best_fitness)
            if (step_count - 1) % 100 == 0:
                self.append(best_fitness, step_count)
        except:
            logging.exception("")

    def summarize(self, agents):
        try:
            logger.debug(self.history)
            logger.debug("best genotype: %s", max(agents, key=lambda a: a.get_fitness()).get_best_genotype())
        except:
            logging.exception("")


class TimeStatistics(StepStatistics):
    def __init__(self, output_file_name='fitness_pyage.txt'):
        super(TimeStatistics, self).__init__(output_file_name)
        self.start = time.time()

    def append(self, best_fitness, step_count):
        self.fitness_output.write(str(time.time() - self.start) + ';' + str(best_fitness) + '\n')



config = None

#TODO refactor
class Inject(object):
    def __init__(self, *args):
        """
        If there are decorator arguments, the function
        to be decorated is not passed to the constructor!
        """
        self.args = args

    def read_config(self, config):
        # about importing modules: http://docs.python.org/2/reference/simple_stmts.html#grammar-token-import_stmt
        exec "import " + config + " as conf"
        return conf

    def __call__(self, f):
        """
        If there are decorator arguments, __call__() is only called
        once, as part of the decoration process!

        """

        def wrapped_f(*args, **kwargs):
            conf = self.read_config(config)
            for arg in self.args:
                conf_arg_name = arg.split(":")[0]
                property_name = arg.split(":")[-1]
                try:
                    attr = getattr(conf, args[0].address.split('.')[0] + '__' + conf_arg_name)()
                except:
                    attr = getattr(conf, conf_arg_name)()
                setattr(args[0], property_name, attr)
            f(*args, **kwargs)

        return wrapped_f


class InjectOptional(Inject):
    def __init__(self, *args):
        super(InjectOptional, self).__init__(*args)

    def __call__(self, f):
        def wrapped_f(*args, **kwargs):
            try:
                conf = self.read_config(config)
                for arg in self.args:
                    conf_arg_name = arg.split(":")[0]
                    property_name = arg.split(":")[-1]
                    try:
                        attr = getattr(conf, args[0].address.split('.')[0] + '__' + conf_arg_name)()
                    except:
                        attr = getattr(conf, conf_arg_name)()
                    setattr(args[0], property_name, attr)
            except:
                pass #parameter is not mandatory
            f(*args, **kwargs)

        return wrapped_f


class Singleton(object):
    def __init__(self, f):
        self.result = None

    def __call__(self, f):
        if not self.result:
            self.result = f()
        return self.result




import logging
from math import sqrt, ceil
import random
import Pyro4
from pyage.core.agent.agent import AGENT
from pyage.core.inject import Inject

logger = logging.getLogger(__name__)


class Locator(object):
    def get_neighbour(self, agent):
        raise NotImplementedError()


class Pyro4Locator(Locator):
    @Inject("ns_hostname")
    def __init__(self):
        super(Pyro4Locator, self).__init__()

    def get_neighbour(self, agent):
        try:
            random_agent = self.__get_random_agent(agent)
            logger.debug(random_agent.get_address())
            return random_agent
        except:
            logging.exception('')

    def __get_random_agent(self, a):
        ns = Pyro4.locateNS(self.ns_hostname)
        agents = ns.list(AGENT)
        logger.debug(agents)
        del agents[AGENT + "." + a.address]
        return Pyro4.Proxy(random.choice(agents.values()))


class RandomLocator(Locator):
    def get_neighbour(self, agent):
        siblings = list(agent.parent.get_agents())
        if len(siblings) < 2:
            return None
        siblings.remove(agent)
        return random.choice(siblings)


class RowLocator(Locator):
    def get_neighbour(self, agent):
        siblings = list(agent.parent.get_agents())
        if len(siblings) < 2:
            return None
        index = siblings.index(agent)
        return random.choice(siblings[index - 2:index] + siblings[index + 1:index + 3])


class GridLocator(Locator):
    def get_neighbour(self, agent):
        siblings = list(agent.parent.get_agents())
        if len(siblings) < 2:
            return None
        index = siblings.index(agent)
        size = len(siblings)
        dim = int(ceil(sqrt(size)))
        return siblings[random.choice([index - dim, index - 1, index + 1, index + dim]) % size]


from pyage.core.address import Addressable
from pyage.core.agent.agent import AbstractAgent
from pyage.core.inject import Inject

class AggregateAgent(Addressable, AbstractAgent):
    @Inject("aggregated_agents:_AggregateAgent__agents")
    def __init__(self, name=None):
        self.name = name
        super(AggregateAgent, self).__init__()
        for agent in self.__agents.values():
            agent.parent = self
        self.steps = 0

    def step(self):
        for agent in self.__agents.values():
            agent.step()

    def remove_agent(self, agent):
        agent = self.__agents[agent.get_address()]
        del self.__agents[agent.get_address()]
        agent.parent = None
        return agent

    def add_agent(self, agent):
        agent.parent = self
        self.__agents[agent.get_address()] = agent

    def get_agents(self):
        return self.__agents.values()

    def get_fitness(self):
        try:
            return max(agent.get_fitness() for agent in self.__agents.values())
        except ValueError:
            return None

    def get_best_genotype(self):
        return max(self.__agents.values(), key=lambda a: a.get_fitness()).get_best_genotype()


def aggregate_agents_factory(*args):
    def factory():
        agents = {}
        for name in args:
            agent = AggregateAgent(name)
            agents[agent.get_address()] = agent
        return agents

    return factory


# coding=utf-8
import logging

from pyage.core import address

from pyage.core.agent.agent import generate_agents, Agent
from pyage.core.locator import RandomLocator
from pyage.core.migration import NoMigration
from pyage.core.statistics import NoStatistics
from pyage.core.stop_condition import StepLimitStopCondition
from pyage.solutions.evolution.crossover import AverageFloatCrossover
from pyage.solutions.evolution.evaluation import FloatRastriginEvaluation
from pyage.solutions.evolution.initializer import FloatInitializer
from pyage.solutions.evolution.mutation import UniformFloatMutation
from pyage.solutions.evolution.selection import TournamentSelection

logger = logging.getLogger(__name__)

agents_count = 1
logger.debug("EVO, %s agents", agents_count)

agents = generate_agents("agent", agents_count, Agent)

stop_condition = lambda: StepLimitStopCondition(10)

size = 500
operators = lambda: [FloatRastriginEvaluation(), TournamentSelection(size=125, tournament_size=125),
                     AverageFloatCrossover(size=size), UniformFloatMutation(probability=0.1, radius=1)]
initializer = lambda: FloatInitializer(2, size, -10, 10)

address_provider = address.SequenceAddressProvider

migration = NoMigration
locator = RandomLocator

stats = NoStatistics



import logging
import os
import urllib2
import Pyro4
import time
import sys
from pyage.core.inject import Inject, InjectOptional
from pyage.core.statistics import Statistics
from pyage.core.workplace import WORKPLACE

logger = logging.getLogger(__name__)


class GlobalStepStatistics(Statistics):
    @Inject("ns_hostname")
    @InjectOptional("notification_url")
    def __init__(self, output_file_name='fitness_pyage.txt'):
        super(GlobalStepStatistics, self).__init__()
        self.fitness_output = open(output_file_name, 'a')
        self.start = time.time()

    def update(self, step_count, agents):
        if step_count % 100 == 0:
            ns = Pyro4.locateNS(self.ns_hostname)
            best_fitness = max(Pyro4.Proxy(w).get_fitness() for w in ns.list(WORKPLACE).values())
            self.append(best_fitness, step_count)


    def append(self, best_fitness, step_count):
        self.fitness_output.write(str(step_count) + ';' + str(abs(best_fitness)) + '\n')

    def summarize(self, agents):
        try:
            if hasattr(self, "notification_url"):
                url = self.notification_url + "?time=%s&agents=%s&conf=%s" % (
                    time.time() - self.start, os.environ['AGENTS'], sys.argv[1])
                logger.info(url)
                urllib2.urlopen(url)

        except:
            logging.exception("")



import logging
import random

from pyage.core.address import Addressable
from pyage.core.inject import Inject


logger = logging.getLogger(__name__)

AGENT = "agent"


class AbstractAgent:
    def initialize(self):
        self.initializer.process(self.population)

    def get_fitness(self):
        return max(genotype.fitness for genotype in self.population)

    def get_best_genotype(self):
        return max(self.population, key=lambda g: g.fitness)

    def add_genotype(self, population):
        logger.debug("received genotype!")
        self.population.extend(population)

    @Inject("operators", "initializer")
    def validate_operators(self):
        for o in self.operators:
            if not o.is_compatible(self.initializer):
                raise ValueError("operator %s is not compatible with %s" % (o, self.initializer))


class Agent(Addressable, AbstractAgent):
    @Inject("locator", "migration")
    def __init__(self, name=None):
        self.name = name
        super(Agent, self).__init__()
        logger.debug("address:" + self.address)
        self.population = []
        self.validate_operators()
        self.initialize()
        self.steps = 0

    def step(self):
        self.steps += 1
        logger.debug("%s %s %s", self.steps, self.address, self.get_fitness())
        for o in self.operators:
            o.process(self.population)
        self.__send_genotype()
        self.__migrate()

    def __send_genotype(self):
        if random.random() < 0.05:
            try:
                neighbour = self.locator.get_neighbour(self)
                if neighbour:
                    logger.debug("neighbour: %s", neighbour.get_address())
                    self.population.sort()
                    sorted_population = list(self.population)
                    sorted_population.sort(key=lambda g: g.fitness)
                    probe = sorted_population[:20:2]
                    neighbour.add_genotype(probe)
            except:
                logging.exception('')

    def __migrate(self):
        self.migration.migrate(self)


def agents_factory(*args):
    def factory():
        agents = {}
        for name in args:
            agent = Agent(name)
            agents[agent.get_address()] = agent
        return agents

    return factory


def generate_agents(prefix, count, type):
    def factory():
        agents = {}
        for i in range(count):
            agent = type(prefix + str(i))
            agents[agent.get_address()] = agent
        return agents

    return factory


def unnamed_agents(count, type):
    def factory():
        agents = {}
        for i in range(count):
            agent = type()
            agents[agent.get_address()] = agent
        return agents

    return factory




import random
from pyage.core.operator import Operator
from pyage.solutions.evolution.genotype import PointGenotype, FloatGenotype

class AbstractMutation(Operator):
    def __init__(self, type, probability):
        super(AbstractMutation, self).__init__()
        self.probability = probability

    def process(self, population):
        for genotype in population:
            if random.random() < self.probability:
                self.mutate(genotype)


class UniformPointMutation(AbstractMutation):
    def __init__(self, probability=0.1, radius=100.5):
        super(UniformPointMutation, self).__init__(PointGenotype, probability)
        self.radius = radius


    def mutate(self, genotype):
        genotype.x = genotype.x + random.uniform(-self.radius, self.radius)
        genotype.y = genotype.y + random.uniform(-self.radius, self.radius)


class UniformFloatMutation(AbstractMutation):
    def __init__(self, probability=0.1, radius=0.5):
        super(UniformFloatMutation, self).__init__(FloatGenotype, probability)
        self.radius = radius


    def mutate(self, genotype):
        index = random.randint(0, len(genotype.genes) - 1)
        genotype.genes[index] += random.uniform(-self.radius, self.radius)




from random import uniform
from pyage.core.emas import EmasAgent
from pyage.core.operator import Operator
from pyage.solutions.evolution.genotype import PointGenotype, FloatGenotype

class PointInitializer(Operator):
    def __init__(self, size=100, lowerbound=0.0, upperbound=1.0):
        super(PointInitializer, self).__init__(PointGenotype)
        self.size = size
        self.lowerbound = lowerbound
        self.upperbound = upperbound

    def process(self, population):
        for i in range(self.size):
            population.append(PointGenotype(self.__randomize(), self.__randomize()))

    def __randomize(self):
        return uniform(self.lowerbound, self.upperbound)


class FloatInitializer(Operator):
    def __init__(self, dims=3, size=100, lowerbound=0.0, upperbound=1.0):
        super(FloatInitializer, self).__init__(FloatGenotype)
        self.size = size
        self.lowerbound = lowerbound
        self.upperbound = upperbound
        self.dims = dims

    def process(self, population):
        for i in range(self.size):
            population.append(FloatGenotype([self.__randomize() for _ in range(self.dims)]))

    def __randomize(self):
        return uniform(self.lowerbound, self.upperbound)

def float_emas_initializer(dims=2,energy=10, size=100, lowerbound=0.0, upperbound=1.0):
    agents = {}
    for i in range(size):
        agent = EmasAgent(FloatGenotype([uniform(lowerbound, upperbound) for _ in range(dims)]), energy)
        agents[agent.get_address()] = agent
    return agents

def emas_initializer(energy=10, size=100, lowerbound=0.0, upperbound=1.0):
    agents = {}
    for i in range(size):
        agent = EmasAgent(PointGenotype(uniform(lowerbound, upperbound), uniform(lowerbound, upperbound)), energy)
        agents[agent.get_address()] = agent
    return agents

